#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import asyncio
import warnings
from enum import Enum
from json import JSONDecoder, JSONEncoder, loads

from .error import MiniZincWarning, error_from_stream_obj
from .types import AnonEnum, ConstrEnum

try:
    import numpy
except ImportError:
    numpy = None


class MZNJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Enum):
            return {"e": o.name}
        if isinstance(o, AnonEnum):
            return {"e": o.enumName, "i": o.value}
        if isinstance(o, ConstrEnum):
            return {"c": o.constructor, "e": o.argument}
        if isinstance(o, set) or isinstance(o, range):
            return {"set": [{"e": i.name} if isinstance(i, Enum) else i for i in o]}
        if numpy is not None:
            if isinstance(o, numpy.ndarray):
                return o.tolist()
            if isinstance(o, numpy.generic):
                return o.item()
        return super().default(o)


class MZNJSONDecoder(JSONDecoder):
    def __init__(self, enum_map=None, *args, **kwargs):
        if enum_map is None:
            self.enum_map = {}
        else:
            self.enum_map = enum_map
        kwargs["object_hook"] = self.mzn_object_hook
        JSONDecoder.__init__(self, *args, **kwargs)

    def transform_enum_object(self, obj):
        # TODO: This probably is an enum, but could still be a record
        if "e" in obj:
            if len(obj) == 1:
                return self.enum_map.get(obj["e"], obj["e"])
            elif len(obj) == 2 and "c" in obj:
                return ConstrEnum(obj["c"], obj["e"])
            elif len(obj) == 2 and "i" in obj:
                return AnonEnum(obj["e"], obj["i"])
        return obj

    def mzn_object_hook(self, obj):
        if isinstance(obj, dict):
            if len(obj) == 1 and "set" in obj:
                if len(obj["set"]) == 1 and isinstance(obj["set"][0], list):
                    assert len(obj["set"][0]) == 2
                    return range(obj["set"][0][0], obj["set"][0][1] + 1)

                li = []
                for item in obj["set"]:
                    if isinstance(item, list):
                        assert len(item) == 2
                        li.extend(list(range(item[0], item[1] + 1)))
                    elif isinstance(item, dict):
                        li.append(self.transform_enum_object(item))
                    else:
                        li.append(item)
                return set(li)
            else:
                return self.transform_enum_object(obj)
        return obj


def decode_json_stream(byte_stream: bytes, cls=None, **kw):
    for line in byte_stream.split(b"\n"):
        line = line.strip()
        if line != b"":
            obj = loads(line, cls=cls, **kw)
            if obj["type"] == "warning" or (
                obj["type"] == "error" and obj["what"] == "warning"
            ):
                # TODO: stack trace and location
                warnings.warn(obj["message"], MiniZincWarning)
            elif obj["type"] == "error":
                raise error_from_stream_obj(obj)
            else:
                yield obj


async def decode_async_json_stream(stream: asyncio.StreamReader, cls=None, **kw):
    buffer: bytes = b""
    while not stream.at_eof():
        try:
            buffer += await stream.readuntil(b"\n")
            buffer = buffer.strip()
            if buffer == b"":
                continue
            obj = loads(buffer, cls=cls, **kw)
            if obj["type"] == "warning" or (
                obj["type"] == "error" and obj["what"] == "warning"
            ):
                # TODO: stack trace and location
                warnings.warn(obj["message"], MiniZincWarning)
            elif obj["type"] == "error":
                raise error_from_stream_obj(obj)
            else:
                yield obj
            buffer = b""
        except asyncio.LimitOverrunError as err:
            buffer += await stream.readexactly(err.consumed)



import sys
from dataclasses import asdict, is_dataclass
from datetime import timedelta
from typing import Any, Dict, Optional, Sequence, Union

import minizinc

if sys.version_info >= (3, 8):
    from typing import Protocol

    class DataClass(Protocol):
        # Checking for this attribute is currently the most reliable way to
        # ascertain that something is a dataclass
        __dataclass_fields__: Dict

else:
    DataClass = Any


def check_result(
    model: minizinc.Model,
    result: minizinc.Result,
    solver: minizinc.Solver,
    solution_nrs: Optional[Sequence[int]] = None,
) -> bool:
    """Checks a result object for a model using the given solver.

    Check the correctness of the solving process using a (different) solver
    configuration. The solver configuration is now used to confirm is
    assignment of the variables is correct. By default only the last solution
    will be checked. A sequence of solution numbers can be provided to check
    multiple solutions.

    Args:
        model (Model): To model for which the solution was provided
        result (Result): The solution to be checked
        solver (Solver): The solver configuration used to check the
            solutions.
        solution_nrs: The index set of solutions to be checked. (default:
            ``-1``)

    Returns:
        bool: True if the given result object is correctly verified.

    """
    if solution_nrs is None:
        solution_nrs = [-1]

    solutions = (
        result.solution if isinstance(result.solution, list) else [result.solution]
    )

    for i in solution_nrs:
        sol = solutions[i]
        if not check_solution(model, sol, result.status, solver):
            return False

    return True


def check_solution(
    model: minizinc.Model,
    solution: Union[DataClass, Dict[str, Any]],
    status: minizinc.Status,
    solver: minizinc.Solver,
) -> bool:
    """Checks a solution for a model using the given solver.

    Check the correctness of the solving process using a (different) solver
    configuration. A new  model instance is created and will be assigned all
    available values from the given solution. The Instance.solve() method is
    then used to ensure that the same solution with the same expected status is
    reached. Note that this method will not check the optimality of a solution.

    Args:
        model (Model): To model for which the solution was provided
        solution (Union[DataClass, Dict[str, Any]]): The solution to be checked
        status (Status): The expected (compatible) MiniZinc status
        solver (Solver): The solver configuration used to check the
            solution.

    Returns:
        bool: True if the given solution are correctly verified.

    """
    instance = minizinc.Instance(solver, model)
    if is_dataclass(solution):
        solution = asdict(solution)

    for k, v in solution.items():
        if k not in ("objective", "__output_item"):
            instance[k] = v
    try:
        check = instance.solve(timeout=timedelta(seconds=5))
        if status == check.status:
            return True
        elif check.status in [
            minizinc.Status.SATISFIED,
            minizinc.Status.OPTIMAL_SOLUTION,
        ] and status in [
            minizinc.Status.SATISFIED,
            minizinc.Status.OPTIMAL_SOLUTION,
            minizinc.Status.ALL_SOLUTIONS,
        ]:
            return True
        else:
            return False
    except minizinc.MiniZincError:
        if status == minizinc.Status.ERROR:
            return True
        return False



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import re
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Tuple


@dataclass
class Location:
    """Representation of a location within a file

    Attributes:
        file (Optional[Path]): Path to the file
        line (int): Line within the file (default: ``0``)
        columns (Tuple[int,int]): Columns on the line, from/to (default:
            ``(0, 0)``)

    """

    file: Optional[Path]
    lines: Tuple[int, int] = (0, 0)
    columns: Tuple[int, int] = (0, 0)


class ConfigurationError(Exception):
    """Exception raised during the configuration of MiniZinc

    Attributes:
        message (str): Explanation of the error
    """

    message: str


class MiniZincError(Exception):
    """Exception raised for errors caused by a MiniZinc Driver

    Attributes:
        location (Optional[Location]): File location of the error
        message (str): Explanation of the error
    """

    location: Optional[Location]
    message: str

    def __init__(self, location: Optional[Location] = None, message: str = ""):
        super().__init__(message)
        self.location = location


class MiniZincWarning(Warning):
    """Warning created for warnings originating from a MiniZinc Driver"""


class EvaluationError(MiniZincError):
    """Exception raised for errors due to an error during instance evaluation by
    the MiniZinc Driver"""

    pass


class AssertionError(EvaluationError):
    """Exception raised for MiniZinc assertions that failed during instance
    evaluation"""

    pass


class TypeError(MiniZincError):
    """Exception raised for type errors found in an MiniZinc Instance"""

    pass


class IncludeError(MiniZincError):
    """Exception raised for type errors found in an MiniZinc Instance"""

    pass


class CyclicIncludeError(MiniZincError):
    """Exception raised for type errors found in an MiniZinc Instance"""

    pass


class SyntaxError(MiniZincError):
    """Exception raised for syntax errors found in an MiniZinc Instance"""

    pass


def parse_error(error_txt: bytes) -> MiniZincError:
    """Parse error from bytes array (raw string)

    Parse error scans the output from a MiniZinc driver to generate the
    appropriate MiniZincError. It will make the distinction between different
    kinds of errors as found by MiniZinc and tries to parse the relevant
    information to the error. The different kinds of errors are represented by
    different sub-classes of MiniZincError.

    Args:
        error_txt (bytes): raw string containing a MiniZinc error. Generally
            this should be the error stream of a driver.

    Returns:
        An error generated from the string

    """
    error = MiniZincError
    if b"MiniZinc: evaluation error:" in error_txt:
        error = EvaluationError
        if b"Assertion failed:" in error_txt:
            error = AssertionError
    elif b"MiniZinc: type error:" in error_txt:
        error = TypeError
    elif b"Error: syntax error" in error_txt:
        error = SyntaxError

    location = None
    match = re.search(rb"([^\s]+):(\d+)(.(\d+)-(\d+))?:\s", error_txt)
    if match:
        columns = (0, 0)
        if match[3]:
            columns = (int(match[4].decode()), int(match[5].decode()))
        lines = (int(match[2].decode()), int(match[2].decode()))
        location = Location(Path(match[1].decode()), lines, columns)

    message = error_txt.decode().strip()
    if not message:
        message = (
            "MiniZinc stopped with a non-zero exit code, but did not output an "
            "error message. "
        )
    elif location is not None and location.file is not None and location.file.exists():
        with location.file.open() as f:
            for _ in range(location.lines[0] - 2):
                f.readline()
            message += "\nFile fragment:\n"
            for nr in range(max(1, location.lines[0] - 1), location.lines[1] + 2):
                line = f.readline()
                if line == "":
                    break
                message += f"{nr}: {line.rstrip()}\n"
                diff = location.columns[1] - location.columns[0]
                if (
                    (location.lines[0] == location.lines[1])
                    and nr == location.lines[0]
                    and diff > 0
                ):
                    message += (
                        " " * (len(str(nr)) + 2 + location.columns[0] - 1)
                        + "^" * (diff + 1)
                        + "\n"
                    )

    return error(location, message)


def error_from_stream_obj(obj):
    """Convert object from JSON stream into MiniZinc Python error

    Args:
        obj (Dict): Parsed JSON object from the ``--json-stream``
            mode of MiniZinc.

    Returns:
        An error generated from the object

    """
    assert obj["type"] == "error"
    error = MiniZincError
    if obj["what"] == "syntax error":
        error = SyntaxError
    elif obj["what"] == "type error":
        error = TypeError
    elif obj["what"] == "include error":
        error = IncludeError
    elif obj["what"] == "cyclic include error":
        error = CyclicIncludeError
    elif obj["what"] == "evaluation error":
        error = EvaluationError
    elif obj["what"] == "assertion failed":
        error = AssertionError

    location = None
    if "location" in obj:
        location = Location(
            obj["location"]["filename"],
            (obj["location"]["firstLine"], obj["location"]["lastLine"]),
            (obj["location"]["firstColumn"], obj["location"]["lastColumn"]),
        )
    # TODO: Process stack information

    message = (
        "MiniZinc stopped with a non-zero exit code, but did not output an "
        "error message. "
    )
    if "message" in obj:
        message = obj["message"]
    elif "cycle" in obj:
        message = " includes ".join(obj["cycle"])

    return error(location, message)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging
import warnings
from typing import Optional

from .driver import Driver
from .error import ConfigurationError, MiniZincError
from .instance import Instance
from .model import Method, Model
from .result import Result, Status
from .solver import Solver
from .types import AnonEnum, ConstrEnum

__version__ = "0.7.0"

logger = logging.getLogger("minizinc")

#: Default MiniZinc driver used by the python package
try:
    default_driver: Optional[Driver] = Driver.find()
    if default_driver is not None:
        default_driver.make_default()
    else:
        warnings.warn(
            "MiniZinc was not found on the system. No default driver could be "
            "initialised.",
            RuntimeWarning,
        )
except ConfigurationError as err:
    warnings.warn(
        f"The MiniZinc version found on the system is incompatible with "
        f"MiniZinc Python:\n{err}\n No default driver could be initialised.",
        RuntimeWarning,
    )

__all__ = [
    "__version__",
    "default_driver",
    "AnonEnum",
    "ConstrEnum",
    "Driver",
    "Instance",
    "Method",
    "MiniZincError",
    "Model",
    "Result",
    "Solver",
    "Status",
]



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import contextlib
import json
import os
import tempfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterator, List, Optional, Tuple

import minizinc


@dataclass
class Solver:
    """The representation of a MiniZinc solver configuration in MiniZinc Python.

    Attributes:
        name (str): The name of the solver.
        version (str): The version of the solver.
        id (str): A unique identifier for the solver, “reverse domain name”
            notation.
        executable (Optional[str]): The executable for this solver that can run
            FlatZinc files. This can be just a file name (in which case the
            solver has to be on the current ``$PATH``), or an absolute path to
            the executable, or a relative path (which is interpreted relative
            to the location of the configuration file). This attribute is set
            to ``None`` if the solver is integrated into MiniZinc.
        mznlib (str): The solver-specific library of global constraints and
            redefinitions. This should be the name of a directory (either an
            absolute path or a relative path, interpreted relative to the
            location of the configuration file). For solvers whose libraries
            are installed in the same location as the MiniZinc standard
            library, this can also take the form -G, e.g., -Ggecode (this is
            mostly the case for solvers that ship with the MiniZinc binary
            distribution).
        mznlibVersion (int): *Currently undocumented in the MiniZinc
            documentation.*
        description (str):  *Currently undocumented in the MiniZinc
            documentation.*
        tags (List[str]): Each solver can have one or more tags that describe
            its features in an abstract way. Tags can be used for selecting a
            solver using the --solver option. There is no fixed list of tags,
            however we recommend using the following tags if they match the
            solver’s behaviour:
            - ``cp``: for Constraint Programming solvers
            - ``mip``: for Mixed Integer Programming solvers
            - ``float``: for solvers that support float variables
            - ``api``: for solvers that use the internal C++ API
        stdFlags (List[str]): Which of the standard solver command line flags
            are supported by this solver. The standard flags are ``-a``,
            ``-n``, ``-s``, ``-v``, ``-p``, ``-r``, ``-f``.
        extraFlags (List[Tuple[str,str,str,str]]): Extra command line flags
            supported by the solver. Each entry must be a tuple of four
            strings. The first string is the name of the option (e.g.
            ``--special-algorithm``). The second string is a description that
            can be used to generate help output (e.g. "which special algorithm
            to use"). The third string specifies the type of the argument
            (``int``, ``bool``, ``float`` or ``string``). The fourth string is
            the default value.
        requiredFlags (List[str]):  *Currently undocumented in the MiniZinc
            documentation.*
        supportsMzn (bool): Whether the solver can run MiniZinc directly (i.e.,
            it implements its own compilation or interpretation of the model).
        supportsFzn (bool): Whether the solver can run FlatZinc. This should be
            the case for most solvers.
        supportsNL (bool): Whether the solver conforms to the AMPL NL standard.
            The NL format is used if ``supportsFZN`` is ``False``.
        needsSolns2Out (bool): Whether the output of the solver needs to be
            passed through the MiniZinc output processor.
        needsMznExecutable (bool): Whether the solver needs to know the
            location of the MiniZinc executable. If true, it will be passed to
            the solver using the ``mzn-executable`` option.
        needsStdlibDir (bool): Whether the solver needs to know the location of
            the MiniZinc standard library directory. If true, it will be passed
            to the solver using the ``stdlib-dir`` option.
        needsPathsFile (bool):  *Currently undocumented in the MiniZinc
            documentation.*
        isGUIApplication (bool): Whether the solver has its own graphical user
            interface, which means that MiniZinc will detach from the process
            and not wait for it to finish or to produce any output.
        _identifier (Optional[str]): A string to specify the solver to MiniZinc
            driver. If set to None, then a solver configuration file should be
            generated.
    """

    name: str
    version: str
    id: str
    executable: Optional[str] = None
    mznlib: str = ""
    mznlibVersion: int = 1
    description: str = ""
    tags: List[str] = field(default_factory=list)
    stdFlags: List[str] = field(default_factory=list)
    extraFlags: List[Tuple[str, str, str, str]] = field(default_factory=list)
    requiredFlags: List[str] = field(default_factory=list)
    inputType: str = "FZN"
    supportsMzn: bool = False
    supportsFzn: bool = True
    supportsNL: bool = False
    needsSolns2Out: bool = False
    needsMznExecutable: bool = False
    needsStdlibDir: bool = False
    needsPathsFile: bool = False
    isGUIApplication: bool = False
    _identifier: Optional[str] = None

    @classmethod
    def lookup(cls, tag: str, driver=None, refresh=False):
        """Lookup a solver configuration in the driver registry.

        Access the MiniZinc driver's known solver configuration and find the
        configuation matching the given tag. Tags are matched in similar to
        ``minizinc --solver tag``. The order of solver configuration attributes
        that are considered is: full id, id ending, tags.

        Args:
            tag (str): tag (or id) of a solver configuration to look up.
            driver (Driver): driver which registry will be searched for the
                solver. If set to None, then ``default_driver`` will be used.
            refresh (bool): Forces the driver to refresh the cache of available solvers.

        Returns:
            Solver: MiniZinc solver configuration compatible with the driver.

        Raises:
            LookupError: No configuration could be located with the given tag.

        """
        if driver is None:
            driver = minizinc.default_driver
        assert driver is not None

        tag_map = driver.available_solvers(refresh)

        if tag not in tag_map or len(tag_map[tag]) < 1:
            raise LookupError(
                f"No solver id or tag '{tag}' found, available options: "
                f"{sorted(tag_map.keys())}"
            )

        return tag_map[tag][0]

    @classmethod
    def load(cls, path: Path):
        """Loads a solver configuration from a file.

        Load solver configuration from a MiniZinc solver configuration given by
        the file on the given location.

        Args:
            path (str): location to the solver configuration file to be loaded.

        Returns:
            Solver: MiniZinc solver configuration compatible with the driver.

        Raises:
            FileNotFoundError: Solver configuration file not found.
            ValueError: File contains an invalid solver configuration.
        """
        if not path.exists():
            raise FileNotFoundError
        solver = json.loads(path.read_bytes())
        # Resolve relative paths
        for key in ["executable", "mznlib"]:
            if key in solver:
                p = Path(solver[key])
                if not p.is_absolute():
                    p = path.parent / p
                    if p.exists():
                        solver[key] = str(p.resolve())

        solver = cls(**solver)
        solver._identifier = str(path.resolve())
        return solver

    @contextlib.contextmanager
    def configuration(self) -> Iterator[str]:
        """Gives the identifier for the current solver configuration.

        Gives an identifier argument that can be used by a CLIDriver to
        identify the solver configuration. If the configuration was loaded
        using the driver and is thus already known, then the identifier will be
        yielded. If the configuration was changed or started from scratch, the
        configuration will be saved to a file and it will yield the name of the
        file.

        Yields:
            str: solver identifier to be used for the ``--solver <id>`` flag.

        """
        try:
            file = None
            if self._identifier is not None:
                yield self._identifier
            else:
                file = tempfile.NamedTemporaryFile(
                    prefix="minizinc_solver_", suffix=".msc", delete=False
                )
                file.write(self.output_configuration().encode())
                file.close()
                yield file.name
        finally:
            if file is not None:
                os.remove(file.name)

    def output_configuration(self) -> str:
        """Formulates a valid JSON specification for the Solver

        Formulates a JSON specification of the solver configuration meant to be
        used by MiniZinc. When stored in a ``.msc`` file it can be used
        directly as a argument to the ``--solver`` flag or stored on the
        MiniZinc solver configuration path. In the latter case it will be
        usable directly from the executable and visible when in ``minizinc
        --solvers``.

        Returns:
            str: JSON string containing the solver specification that can be
                read by MiniZinc

        """
        return json.dumps(
            {
                # TODO: Output inputType flag when fully supported
                "name": self.name,
                "version": self.version,
                "id": self.id,
                "executable": self.executable,
                "mznlib": self.mznlib,
                "tags": self.tags,
                "stdFlags": self.stdFlags,
                "extraFlags": self.extraFlags,
                "supportsMzn": self.supportsMzn,
                "supportsFzn": self.supportsFzn,
                "needsSolns2Out": self.needsSolns2Out,
                "needsMznExecutable": self.needsMznExecutable,
                "needsStdlibDir": self.needsStdlibDir,
                "isGUIApplication": self.isGUIApplication,
            },
            sort_keys=True,
            indent=4,
        )

    def __setattr__(self, key, value):
        if (
            key
            in [
                "version",
                "executable",
                "mznlib",
                "tags",
                "stdFlags",
                "extraFlags",
                "inputType",
                "supportsMzn",
                "supportsFzn",
                "needsSolns2Out",
                "needsMznExecutable",
                "needsStdlibDir",
                "isGUIApplication",
            ]
            and getattr(self, key, None) is not value
        ):
            self._identifier = None
        return super().__setattr__(key, value)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['minizinc']

package_data = \
{'': ['*']}

extras_require = \
{'dzn': ['lark-parser>=0.12.0,<0.13.0']}

entry_points = \
{'pygments.lexers': ['minizinclexer = minizinc.pygments:MiniZincLexer']}

setup_kwargs = {
    'name': 'minizinc',
    'version': '0.8.0',
    'description': 'Access MiniZinc directly from Python',
    'long_description': '<!-- PROJECT LOGO -->\n<br />\n<p align="center">\n  <a href="https://github.com/MiniZinc/minizinc-python">\n    <img src="https://www.minizinc.org/MiniZn_logo.png" alt="Logo" width="80" height="80">\n  </a>\n\n  <h3 align="center">MiniZinc Python</h3>\n\n  <p align="center">\n    The python package that allows you to access all of MiniZinc\'s functionalities directly from Python.\n    <br />\n    <a href="https://minizinc-python.readthedocs.io/en/latest/"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    <a href="https://github.com/MiniZinc/minizinc-python/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/MiniZinc/minizinc-python/issues">Request Feature</a>\n  </p>\n</p>\n\n\n<!-- TABLE OF CONTENTS -->\n## Table of Contents\n\n* [About the Project](#about-the-project)\n* [Getting Started](#getting-started)\n  * [Installation](#installation)\n  * [Usage](#usage)\n* [Testing](#testing)\n* [Roadmap](#roadmap)\n* [Contributing](#contributing)\n* [License](#license)\n* [Contact](#contact)\n<!-- * [Acknowledgements](#acknowledgements) -->\n\n\n<!-- ABOUT THE PROJECT -->\n## About The Project\n\n_MiniZinc Python_ provides an interface from Python to the MiniZinc driver. The\nmost important goal of this project are to allow easy access to MiniZinc using\nnative Python structures. This will allow you to more easily make scripts to run\nMiniZinc, but will also allow the integration of MiniZinc models within bigger\n(Python) projects. This module also aims to expose an interface for meta-search.\nFor problems that are hard to solve, meta-search can provide solutions to reach\nmore or better solutions quickly.\n\n\n<!-- GETTING STARTED -->\n## Getting Started\n\nTo get a MiniZinc Python up and running follow these simple steps.\n\n### Installation\n\n_MiniZinc Python_ can be installed by running `pip install minizinc`. It\nrequires [MiniZinc](https://www.minizinc.org/) 2.5.0+ and\n[Python](https://www.python.org/) 3.6.0+ to be installed on the system. MiniZinc\npython expects the `minizinc` executable to be available on the executable path,\nthe `$PATH` environmental variable, or in a default installation location.\n\n_For more information, please refer to the\n[Documentation](https://minizinc-python.readthedocs.io/en/latest/)_\n\n\n### Usage\n\nOnce all prerequisites and MiniZinc Python are installed, a `minizinc` module\nwill be available in Python. The following Python code shows how to run a\ntypical MiniZinc model.\n\n```python\nimport minizinc\n\n# Create a MiniZinc model\nmodel = minizinc.Model()\nmodel.add_string("""\nvar -100..100: x;\nint: a; int: b; int: c;\nconstraint a*(x*x) + b*x = c;\nsolve satisfy;\n""")\n\n# Transform Model into a instance\ngecode = minizinc.Solver.lookup("gecode")\ninst = minizinc.Instance(gecode, model)\ninst["a"] = 1\ninst["b"] = 4\ninst["c"] = 0\n\n# Solve the instance\nresult = inst.solve(all_solutions=True)\nfor i in range(len(result)):\n    print("x = {}".format(result[i, "x"]))\n```\n\n_For more examples, please refer to the\n[Documentation](https://minizinc-python.readthedocs.io/en/latest/)_\n\n<!-- TESTING INSTRUCTIONS -->\n## Testing\n\nMiniZinc Python uses [Tox](https://pypi.org/project/tox/) environments to test\nits coding style and functionality. The code style tests are executed using\n[Black](https://pypi.org/project/black/),\n[Flake8](https://pypi.org/project/flake8/), and\n[isort](https://pypi.org/project/isort/). The functionality tests are\nconstructed using the [PyTest]() unit testing framework.\n\n  * To run all tests, simply execute `tox` in the repository directory.\n  * Individual environments can be triggered using the `-e` flag.\n    * To test the coding style of the repository run `tox -e check`\n    * The `py3x` environments are used to test a specific Python version; for\n      example, to test using Python version 3.7 run `tox -e py37`\n\nTox can also be used to generate the documentation, `tox -e docs`, and to\ntypeset the Python code, `tox -e format`.\n\n<!-- ROADMAP -->\n## Roadmap\n\nSee the [open issues](https://github.com/MiniZinc/minizinc-python/issues) for a\nlist of proposed features (and known issues).\n\n\n<!-- CONTRIBUTING -->\n## Contributing\n\nContributions are what make the open source community such an amazing place to\nbe learn, inspire, and create. Any contributions you make are **greatly\nappreciated**.\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n\n\n<!-- LICENSE -->\n## License\n\nDistributed under the Mozilla Public License Version 2.0. See `LICENSE` for more information.\n\n\n<!-- CONTACT -->\n## Contact\n👤 **Jip J. Dekker**\n  * Twitter: [@DekkerOne](https://twitter.com/DekkerOne)\n  * Github: [Dekker1](https://github.com/Dekker1)\n\n🏛 **MiniZinc**\n  * Website: [https://www.minizinc.org/](https://www.minizinc.org/)\n\n<!-- ACKNOWLEDGEMENTS -->\n<!-- ## Acknowledgements -->\n\n<!-- * []() -->\n<!-- * []() -->\n<!-- * []() -->\n',
    'author': 'Jip J. Dekker',
    'author_email': 'jip@dekker.one',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.minizinc.org/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.
import asyncio
import contextlib
import json
import os
import re
import sys
import tempfile
import warnings
from dataclasses import field, make_dataclass
from datetime import timedelta
from enum import EnumMeta
from keyword import iskeyword
from numbers import Number
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

import minizinc

from .driver import Driver
from .error import MiniZincError, parse_error
from .json import (
    MZNJSONDecoder,
    MZNJSONEncoder,
    decode_async_json_stream,
    decode_json_stream,
)
from .model import Method, Model, ParPath, UnknownExpression
from .result import Result, Status, parse_solution, set_stat
from .solver import Solver

if sys.version_info >= (3, 8):
    from typing import Final

    SEPARATOR: Final[bytes] = str.encode("----------" + os.linesep)
else:
    SEPARATOR: bytes = str.encode("----------" + os.linesep)


class _GeneratedSolution:
    pass


class Instance(Model):
    """Representation of a MiniZinc instance in Python.

    Raises:
        MiniZincError: when an error occurs during the parsing or
            type checking of the model object.
    """

    _driver: Driver
    _solver: Solver
    _input_cache: Optional[Dict[str, Type]] = None
    _output_cache: Optional[Dict[str, Type]] = None
    _method_cache: Optional[Method] = None
    _parent: Optional["Instance"] = None
    _field_renames: List[Tuple[str, str]]

    def __init__(
        self,
        solver: Solver,
        model: Optional[Model] = None,
        driver: Optional[Driver] = None,
    ):
        super().__init__()
        self._solver = solver
        self._field_renames = []
        if driver is not None:
            self._driver = driver
        elif minizinc.default_driver is not None:
            self._driver = minizinc.default_driver
        else:
            raise Exception("No compatible driver provided")

        if model is not None:
            self.output_type = model.output_type
            self._includes = model._includes.copy()
            self._code_fragments = model._code_fragments.copy()
            self._data = dict.copy(model._data)
            self._enum_map = dict.copy(model._enum_map)
            self._checker = True

            # Generate output_type to ensure the same type between different
            # instances of the same model
            if self.output_type is None:
                self.analyse()
                model.output_type = self.output_type

    def solve(
        self,
        timeout: Optional[timedelta] = None,
        nr_solutions: Optional[int] = None,
        processes: Optional[int] = None,
        random_seed: Optional[int] = None,
        all_solutions: bool = False,
        intermediate_solutions: bool = False,
        free_search: bool = False,
        optimisation_level: Optional[int] = None,
        **kwargs,
    ) -> Result:
        """Solves the Instance using its given solver configuration.

        Find the solutions to the given MiniZinc instance using the given solver
        configuration. First, the Instance will be ensured to be in a state
        where the solver specified in the solver configuration can understand
        the problem and then the solver will be requested to find the
        appropriate solution(s) to the problem.

        Args:
            timeout (Optional[timedelta]): Set the time limit for the process of
                solving the instance.
            nr_solutions (Optional[int]): The requested number of solution.
                (Only available on satisfaction problems and when the ``-n``
                flag is supported by the solver).
            processes (Optional[int]): Set the number of processes the solver
                can use. (Only available when the ``-p`` flag is supported by
                the solver).
            random_seed (Optional[int]): Set the random seed for solver. (Only
                available when the ``-r`` flag is supported by the solver).
            free_search (bool): Allow the solver to ignore the search definition
                within the instance. (Only available when the ``-f`` flag is
                supported by the solver).
            all_solutions (bool): Request to solver to find all solutions. (Only
                available on satisfaction problems and when the ``-a`` flag is
                supported by the solver)
            intermediate_solutions (bool): Request the solver to output any
                intermediate solutions that are found during the solving
                process. (Only available on optimisation problems and when the
                ``-a`` flag is supported by the solver)
            optimisation_level (Optional[int]): Set the MiniZinc compiler
                optimisation level.

                - 0: Disable optimisation
                - 1: Single pass optimisation (default)
                - 2: Flatten twice to improve flattening decisions
                - 3: Perform root-node-propagation
                - 4: Probe bounds of all variables at the root node
                - 5: Probe values of all variables at the root node

            **kwargs: Other flags to be passed onto the solver. ``--`` can be
                omitted in the name of the flag. If the type of the flag is
                Boolean, then its value signifies its occurrence.

        Returns:
            Tuple[Status, Optional[Union[List[Dict], Dict]], Dict]:
                tuple containing solving status, values assigned in the
                solution, and statistical information. If no solutions is found
                the second member of the tuple is ``None``.

        Raises:
            MiniZincError: An error occurred while compiling or solving the
                model instance.

        """
        coroutine = self.solve_async(
            timeout=timeout,
            nr_solutions=nr_solutions,
            processes=processes,
            random_seed=random_seed,
            all_solutions=all_solutions,
            intermediate_solutions=intermediate_solutions,
            free_search=free_search,
            optimisation_level=optimisation_level,
            **kwargs,
        )
        try:
            if sys.platform == "win32":
                asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
            return asyncio.run(coroutine)
        except RuntimeError as r:
            if "called from a running event loop" in r.args[0]:
                raise RuntimeError(
                    "the synchronous MiniZinc Python `solve()` method was called from"
                    " an asynchronous environment.\n\nBecause Python's asyncio library"
                    " does not support using multiple event loops that would be"
                    " required to use this method, we instead suggest you use the"
                    " `solve_async()' method or patch Python behaviour with a package"
                    " such as `nested_asyncio'.\n\nOriginal message: " + str(r)
                ) from r
            else:
                raise r

    async def solve_async(
        self,
        timeout: Optional[timedelta] = None,
        nr_solutions: Optional[int] = None,
        processes: Optional[int] = None,
        random_seed: Optional[int] = None,
        all_solutions=False,
        intermediate_solutions=False,
        free_search: bool = False,
        optimisation_level: Optional[int] = None,
        **kwargs,
    ) -> Result:
        """Solves the Instance using its given solver configuration in a coroutine.

        This method returns a coroutine that finds solutions to the given
        MiniZinc instance. For more information regarding this methods and its
        arguments, see the documentation of :func:`~MiniZinc.Instance.solve`.

        Returns:
            Tuple[Status, Optional[Union[List[Dict], Dict]], Dict]:
                tuple containing solving status, values assigned, and
                statistical information.

        Raises:
            MiniZincError: An error occurred while compiling or solving the
                model instance.

        """
        status = Status.UNKNOWN
        solution = None
        statistics: Dict[str, Any] = {}

        multiple_solutions = (
            all_solutions or intermediate_solutions or nr_solutions is not None
        )
        if multiple_solutions:
            solution = []

        async for result in self.solutions(
            timeout=timeout,
            nr_solutions=nr_solutions,
            processes=processes,
            random_seed=random_seed,
            all_solutions=all_solutions,
            free_search=free_search,
            optimisation_level=optimisation_level,
            **kwargs,
        ):
            status = result.status
            statistics.update(result.statistics)
            if result.solution is not None:
                if multiple_solutions:
                    solution.append(result.solution)
                else:
                    solution = result.solution
        return Result(status, solution, statistics)

    async def solutions(
        self,
        timeout: Optional[timedelta] = None,
        nr_solutions: Optional[int] = None,
        processes: Optional[int] = None,
        random_seed: Optional[int] = None,
        all_solutions=False,
        intermediate_solutions=False,
        free_search: bool = False,
        optimisation_level: Optional[int] = None,
        verbose: bool = False,
        debug_output: Optional[Path] = None,
        **kwargs,
    ) -> AsyncIterator[Result]:
        """An asynchronous generator for solutions of the MiniZinc instance.

        This method provides an asynchronous generator for the solutions of the
        MiniZinc instance. Every (intermediate) solution is yielded one at a
        time, the last item yielded from the generator will not contain a new
        solution, but will return the final Status and all remaining
        statistical values. For more information regarding this methods and its
        arguments, see the documentation of :func:`~MiniZinc.Instance.solve`.

        Yields:
            Result:
                A Result object containing the current solving status, values
                assigned, and statistical information.

        """
        method = self.method  # Ensure self.analyse() has been executed
        # Set standard command line arguments
        cmd: List[Any] = [
            "--output-mode",
            "json",
            "--output-time",
            "--output-objective",
            "--output-output-item",
            "--statistics",  # Enable statistics
            # Enable intermediate solutions
            # (ensure that solvers always output their best solution)
            "--intermediate-solutions",
        ]

        # Process number of solutions to be generated
        if all_solutions:
            if nr_solutions is not None:
                raise ValueError(
                    "The number of solutions cannot be limited when looking "
                    "for all solutions"
                )
            if method != Method.SATISFY:
                raise NotImplementedError(
                    "Finding all optimal solutions is not yet implemented"
                )
            if "-a" not in self._solver.stdFlags:
                raise NotImplementedError("Solver does not support the -a flag")
            cmd.append("--all-solutions")
        elif nr_solutions is not None:
            if nr_solutions <= 0:
                raise ValueError(
                    "The number of solutions can only be set to a positive "
                    "integer number"
                )
            if self.method != Method.SATISFY:
                raise NotImplementedError(
                    "Finding multiple optimal solutions is not yet implemented"
                )
            if "-n" not in self._solver.stdFlags:
                raise NotImplementedError("Solver does not support the -n flag")
            cmd.extend(["--num-solutions", str(nr_solutions)])
        # Set number of processes to be used
        if processes is not None:
            cmd.extend(["--parallel", str(processes)])
        # Set random seed to be used
        if random_seed is not None:
            cmd.extend(["--random-seed", str(random_seed)])
        # Enable free search if specified
        if free_search:
            cmd.append("--free-search")
        # Set compiler optimisation level if specified
        if optimisation_level is not None:
            cmd.extend(["-O", str(optimisation_level)])

        # Set time limit for the MiniZinc solving
        if timeout is not None:
            cmd.extend(["--time-limit", str(int(timeout.total_seconds() * 1000))])

        if verbose:
            cmd.append("--verbose")

        for flag, value in kwargs.items():
            if not flag.startswith("-"):
                flag = "--" + flag
            if type(value) is bool:
                if value:
                    cmd.append(flag)
            else:
                cmd.extend([flag, value])

        # Add files as last arguments
        with self.files() as files, self._solver.configuration() as solver:
            assert self.output_type is not None
            cmd.extend(files)

            status = Status.UNKNOWN
            last_status = Status.UNKNOWN
            code = 0
            statistics: Dict[str, Any] = {}

            try:
                # Run the MiniZinc process
                proc = await self._driver._create_process(cmd, solver=solver)
                assert isinstance(proc.stderr, asyncio.StreamReader)
                assert isinstance(proc.stdout, asyncio.StreamReader)

                # Python 3.7+: replace with asyncio.create_task
                read_stderr = asyncio.ensure_future(_read_all(proc.stderr))

                if self._driver.parsed_version >= (2, 6, 0):
                    async for obj in decode_async_json_stream(
                        proc.stdout, cls=MZNJSONDecoder, enum_map=self._enum_map
                    ):
                        solution, new_status, statistics = self._parse_stream_obj(
                            obj, statistics
                        )
                        if new_status is not None:
                            status = new_status
                        elif solution is not None:
                            if status == Status.UNKNOWN:
                                status = Status.SATISFIED
                            yield Result(status, solution, statistics)
                            last_status = status
                            solution = None
                            statistics = {}
                else:
                    async for raw_sol in _seperate_solutions(proc.stdout):
                        status = Status.SATISFIED
                        solution, statistics = parse_solution(
                            raw_sol,
                            self.output_type,
                            self._enum_map,
                            self._field_renames,
                        )
                        yield Result(Status.SATISFIED, solution, statistics)

                code = await proc.wait()
            except asyncio.IncompleteReadError as err:
                # End of Stream has been reached
                # Read remaining text in buffer
                code = await proc.wait()
                remainder = err.partial

                # Parse and output the remaining statistics and status messages
                if self._driver.parsed_version >= (2, 6, 0):
                    for obj in decode_json_stream(
                        remainder, cls=MZNJSONDecoder, enum_map=self._enum_map
                    ):
                        solution, new_status, statistics = self._parse_stream_obj(
                            obj, statistics
                        )
                        if new_status is not None:
                            status = new_status
                        elif solution is not None:
                            if status == Status.UNKNOWN:
                                status = Status.SATISFIED
                            yield Result(status, solution, statistics)
                            solution = None
                            statistics = {}
                else:
                    for res in filter(None, remainder.split(SEPARATOR)):
                        new_status = Status.from_output(res, method)
                        if new_status is not None:
                            status = new_status
                        solution, statistics = parse_solution(
                            res,
                            self.output_type,
                            self._enum_map,
                            self._field_renames,
                        )
                        yield Result(status, solution, statistics)
            except (asyncio.CancelledError, MiniZincError, Exception):
                # Process was cancelled by the user, a MiniZincError occurred, or
                # an unexpected Python exception occurred
                # First, terminate the process
                proc.terminate()
                _ = await proc.wait()
                # Then, reraise the error that occurred
                raise
            if self._driver.parsed_version >= (2, 6, 0) and (
                status != last_status or statistics != {}
            ):
                yield Result(status, None, statistics)

            # Raise error if required
            stderr = await read_stderr
            if code != 0 or status == Status.ERROR:
                raise parse_error(stderr)

            if debug_output is not None:
                debug_output.write_bytes(stderr)

    @contextlib.contextmanager
    def branch(self) -> Iterator["Instance"]:  # TODO: Self reference
        """Create a branch of the current instance

        Branches from the current instance and yields a child instance. Any
        changes made to the child instance can not influence the current
        instance. WARNING: The branch method assumes that no changes will be
        made to the parent method while the child instance is still alive.
        Changes to the parent model are locked until the child method are
        destroyed.

        Yields:
            Instance: branched child instance
        """
        child = self.__class__(self._solver)
        child._parent = self

        # Copy current information from analysis
        child._method_cache = self.method
        child.output_type = self.output_type
        child._output_cache = self._output_cache
        child._input_cache = self._input_cache

        with self._lock:
            yield child

    @contextlib.contextmanager
    def files(self) -> Iterator[List[Path]]:
        """Gets list of files of the Instance

        Files will create a list of paths to the files that together form the
        Instance. Parts of the Instance might be saved to files and are only
        guaranteed to exist while within the created context.

        Yields:
            List of Path objects to existing and created files
        """

        files: List[Path] = []
        fragments: List[str] = []
        data: Dict[str, Any] = {}

        inst: Optional["Instance"] = self
        while inst is not None:
            for k, v in inst._data.items():
                if isinstance(v, UnknownExpression) or k in data:
                    fragments.append(f"{k} = {v};\n")
                elif isinstance(v, EnumMeta):
                    enumDecl = k + " = {"
                    first = True
                    for mem in v.__members__:
                        if not first:
                            enumDecl += ", "
                        enumDecl += "'" + str(mem) + "'"
                        first = False
                    enumDecl += "};"
                    fragments.append(enumDecl)
                else:
                    data[k] = v
            fragments.extend(inst._code_fragments)
            files.extend(inst._includes)

            inst = inst._parent

        gen_files = []
        try:
            if len(data) > 0:
                file = tempfile.NamedTemporaryFile(
                    prefix="mzn_data", suffix=".json", delete=False
                )
                gen_files.append(file)
                file.write(json.dumps(data, cls=MZNJSONEncoder).encode())
                file.close()
                files.append(Path(file.name))
            if len(fragments) > 0 or len(files) == 0:
                file = tempfile.NamedTemporaryFile(
                    prefix="mzn_fragment", suffix=".mzn", delete=False
                )
                gen_files.append(file)
                for code in fragments:
                    file.write(code.encode())
                file.close()
                files.append(Path(file.name))
            yield files
        finally:
            for file in gen_files:
                file.close()
                os.remove(file.name)

    @property
    def method(self) -> Method:
        """Query the Method used by the Instance.

        Returns:
            Method: Method of the goal used by the Instance.
        """
        if self._method_cache is None:
            self.analyse()
        assert self._method_cache is not None
        return self._method_cache

    @property
    def input(self) -> Dict[str, Type]:
        """Query the input parameters of the Instance.

        Returns:
            Dict[str, Type]: A mapping from parameter identifiers to their Types.
        """
        if self._input_cache is None or self._method_cache is None:
            self.analyse()
        assert self._input_cache is not None
        return self._input_cache

    @property
    def output(self):
        """Query the output parameters of the Instance.

        Returns:
            Dict[str, Type]: A mapping from parameter identifiers to their Types.
        """
        if self._output_cache is None or self._method_cache is None:
            self.analyse()
        assert self._output_cache is not None
        return self._output_cache

    def analyse(self):
        """Discovers basic information about a CLIInstance

        Analyses a given instance and discovers basic information about set
        model such as the solving method, the input parameters, and the output
        parameters. The information found will be stored among the attributes
        of the instance.
        """
        with self.files() as files:
            assert len(files) > 0
            output = self._driver._run(["--model-interface-only"] + files, self._solver)
        if self._driver.parsed_version >= (2, 6, 0):
            interface = None
            for obj in decode_json_stream(output.stdout):
                if obj["type"] == "interface":
                    interface = obj
                    break
        else:
            interface = json.loads(output.stdout)
        old_method = self._method_cache
        self._method_cache = Method.from_string(interface["method"])
        self._input_cache = {}
        for key, value in interface["input"].items():
            self._input_cache[key] = _to_python_type(value)
        old_output = self._output_cache
        self._output_cache = {}
        for key, value in interface["output"].items():
            self._output_cache[key] = _to_python_type(value)
        if interface.get("has_output_item", True):
            self._output_cache["_output_item"] = str
        if self._checker:
            self._output_cache["_checker"] = str

        if self.output_type is None or (
            issubclass(self.output_type, _GeneratedSolution)
            and (self._output_cache != old_output or self._method_cache != old_method)
        ):
            fields = []
            self._field_renames = []
            if (
                self._method_cache is not Method.SATISFY
                and "objective" not in self._output_cache
            ):
                fields.append(("objective", Number))
            for k, v in self._output_cache.items():
                if k in ["_output_item", "_checker"]:
                    fields.append((k, str, field(default="")))
                elif iskeyword(k):
                    warnings.warn(
                        f"MiniZinc field '{k}' is a Python keyword. It has been "
                        f"renamed to 'mzn_{k}'",
                        SyntaxWarning,
                    )
                    self._field_renames.append((k, "mzn_" + k))
                    fields.append(("mzn_" + k, v))
                else:
                    fields.append((k, v))

            minizinc.logger.debug(
                f"CLIInstance:analyse -> output fields: " f"{[f[0:2] for f in fields]}"
            )

            methods = {}
            if interface.get("has_output_item", True):
                methods["__str__"] = (
                    lambda myself: myself.__repr__()
                    if myself._output_item == ""
                    else myself._output_item
                )
            if self._checker:
                methods["check"] = lambda myself: myself._checker

            self.output_type = make_dataclass(
                "Solution",
                fields,
                bases=(_GeneratedSolution,),
                namespace=methods,
                frozen=True,
            )

    def _reset_analysis(self):
        self._method_cache = None

    @contextlib.contextmanager
    def flat(
        self,
        timeout: Optional[timedelta] = None,
        optimisation_level: Optional[int] = None,
        **kwargs,
    ):
        """Produce a FlatZinc file for the instance.

        Args:
            timeout (Optional[timedelta]): Set the time limit for the process
                of flattening the instance. TODO: An exception is raised if the
                timeout is reached.
            optimisation_level (Optional[int]): Set the MiniZinc compiler
                optimisation level.

                - 0: Disable optimisation
                - 1: Single pass optimisation (default)
                - 2: Flatten twice to improve flattening decisions
                - 3: Perform root-node-propagation
                - 4: Probe bounds of all variables at the root node
                - 5: Probe values of all variables at the root node

            **kwargs: Other flags to be passed to the compiler. ``--`` can be
                omitted in the name of the flag. If the type of the flag is
                Boolean, then its value signifies its occurrence.

        Yields:
            Tuple containing the files of the FlatZinc model, the output model
            and a dictionary the statistics of flattening

        """
        cmd: List[Any] = ["--compile", "--statistics"]

        fzn = tempfile.NamedTemporaryFile(prefix="fzn_", suffix=".fzn", delete=False)
        cmd.extend(["--fzn", fzn.name])
        fzn.close()
        ozn = tempfile.NamedTemporaryFile(prefix="ozn_", suffix=".fzn", delete=False)
        cmd.extend(["--ozn", ozn.name])
        ozn.close()

        if timeout is not None:
            cmd.extend(["--time-limit", str(int(timeout.total_seconds() * 1000))])

        # Set compiler optimisation level if specified
        if optimisation_level is not None:
            cmd.extend(["-O", str(optimisation_level)])

        for flag, value in kwargs.items():
            if not flag.startswith("-"):
                flag = "--" + flag
            if type(value) is bool:
                if value:
                    cmd.append(flag)
            else:
                cmd.extend([flag, value])

        # Add files as last arguments
        with self.files() as files:
            cmd.extend(files)
            # Run the MiniZinc process
            output = self._driver._run(cmd, solver=self._solver)

        statistics: Dict[str, Any] = {}
        matches = re.findall(rb"%%%mzn-stat:? (\w*)=([^\r\n]*)", output.stdout)
        for m in matches:
            set_stat(statistics, m[0].decode(), m[1].decode())

        try:
            yield fzn, ozn, statistics
        finally:
            os.remove(fzn.name)
            os.remove(ozn.name)

    def add_file(self, file: ParPath, parse_data: bool = True) -> None:
        self._reset_analysis()
        return super().add_file(file, parse_data)

    def add_string(self, code: str) -> None:
        self._reset_analysis()
        return super().add_string(code)

    def _parse_stream_obj(self, obj, statistics):
        solution = None
        status = None
        if obj["type"] == "solution":
            tmp = obj["output"]["json"]
            if "_objective" in tmp:
                tmp["objective"] = tmp.pop("_objective")
            if "_output" in tmp:
                tmp["_output_item"] = tmp.pop("_output")
            for before, after in self._field_renames:
                tmp[after] = tmp.pop(before)

            if "_checker" in statistics:
                tmp["_checker"] = statistics.pop("_checker")

            solution = self.output_type(**tmp)
            statistics["time"] = obj["time"]
        elif obj["type"] == "time":
            statistics["time"] = obj["time"]
        elif obj["type"] == "statistics":
            for key, val in obj["statistics"].items():
                set_stat(statistics, key, str(val))
        elif obj["type"] == "status":
            status = Status.from_str(obj["status"])
        elif obj["type"] == "checker":
            if "raw" in obj["output"]:
                statistics["_checker"] = obj["output"]["raw"]
            else:
                # TODO: can we ensure this is made JSON?
                statistics["_checker"] = obj["output"]["dzn"]

        return solution, status, statistics


def _to_python_type(mzn_type: dict) -> Type:
    """Converts MiniZinc JSON type to Type

    Converts a MiniZinc JSON type definition generated by the MiniZinc CLI to a
    Python Type object. This can be used on types that result from calling
    ``minizinc --model-interface-only``.

    Args:
        mzn_type (dict): MiniZinc type definition as resulting from JSON

    Returns:
        Type: Type definition in Python

    """
    basetype = mzn_type["type"]
    pytype: Type
    # TODO: MiniZinc does not report enumerated types correctly
    if basetype == "bool":
        pytype = bool
    elif basetype == "float":
        pytype = float
    elif basetype == "int":
        pytype = int
    elif basetype == "string":
        pytype = str
    elif basetype == "ann":
        pytype = str
    else:
        warnings.warn(
            f"Unable to determine minizinc type `{basetype}` assuming integer type",
            FutureWarning,
        )
        pytype = int

    if mzn_type.get("set", False):
        if pytype is int:
            pytype = Union[Set[int], range]  # type: ignore
        else:
            pytype = Set[pytype]  # type: ignore

    dim = mzn_type.get("dim", 0)
    while dim >= 1:
        # No typing support for n-dimensional typing
        pytype = List[pytype]  # type: ignore
        dim -= 1
    return pytype


async def _seperate_solutions(stream: asyncio.StreamReader):
    solution: bytes = b""
    while not stream.at_eof():
        try:
            solution += await stream.readuntil(SEPARATOR)
            yield solution
            solution = b""
        except asyncio.LimitOverrunError as err:
            solution += await stream.readexactly(err.consumed)


async def _read_all(stream: asyncio.StreamReader):
    output: bytes = b""
    while not stream.at_eof():
        try:
            output += await stream.read()
            return output
        except asyncio.LimitOverrunError as err:
            output += await stream.readexactly(err.consumed)
    return output
